#include <windows.h>
#include <gl/gl.h>
#include <gl/glaux.h>
#include <time.h>

/* global variables */
bool g_keyboardInput[256];

int g_paddleCoord;
int	g_paddleHorizontalCoord;
int g_changePos;

int g_ballXCoord;
int g_ballYCoord;

int g_ballMoveX;
int g_ballMoveY;

UINT ballVelocity;

bool *g_blockVisible;
UINT g_blockNumber;

DWORD dwExtendedStyle;
DWORD dwWindowStyle;
WNDCLASS windowClassStruct;

double GetRandomDouble(double min, double max);
void BlockRenderer(int x_coord, int y_coord, int width, int height);
void ProjectileRenderer(int x_coord, int y_coord, int width, int height);

int ProjectilePaddleCollision(int position_y, int variant);

HDC hDeviceContext = NULL;
HGLRC hRenderContext = NULL;
HWND hWindow = NULL;
HINSTANCE hInstance;

UINT g_programRunning = 1;

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

struct ObjectMeasurement {
	int width;
	int height;
} size;

/* generic function that returns a random double value between the provided bounds */
double GetRandomDouble(double Low, double High)
{
	return ((double) rand() / (RAND_MAX+1.0)) * (High-Low) + Low;
}

/* main block rendering function */
void RenderSingleBlock(float x_coord1, float y_coord1, float x_coord2, float y_coord2)
{
	glColor3f((float)GetRandomDouble(80,256)/256.0f, (float)GetRandomDouble(80,256)/256.0f, (float)GetRandomDouble(80,256)/256.0f);
	
	glBegin(GL_QUADS);
		glVertex2f(x_coord1,y_coord1);
		glVertex2f(x_coord2,y_coord1);
		glVertex2f(x_coord2,y_coord2);
		glVertex2f(x_coord1,y_coord2);
	glEnd();
}

/* renders each block */
void BlockRenderer(int x_coord, int y_coord, int width, int height)
{
	float x_coord1,y_coord1,x_coord2,y_coord2;
	x_coord1=float(x_coord-160)*5.0f/320.0f;
	y_coord1=float(100-y_coord)*4.2f/200.0f;
	x_coord2=float(x_coord+width-160)*5.0f/320.0f;
	y_coord2=float(100-y_coord-height)*4.2f/200.0f;
	RenderSingleBlock(x_coord1,y_coord1,x_coord2,y_coord2);
}

/* render the projectile */
void ProjectileRenderer(int x_coord, int y_coord, int width, int height)
{
	float x_coord1, x_coord2;
	float y_coord1, y_coord2;

	x_coord1 = (x_coord - 150)*0.016f;
	x_coord2 = (x_coord + width - 150)*0.016f;
	
	y_coord1 = (90 - y_coord)*0.021f;
	y_coord2 = (90 - y_coord-height)*0.021f;
	
	RenderSingleBlock(x_coord1,y_coord1,x_coord2,y_coord2);
}

/* detect collisions between the projectile and paddle */
int ProjectilePaddleCollision(int position_y, int variant)
{
    int orientation = -1;
    int count = -3, incr = 0;

    if (position_y > 0)
      orientation = 1;
    
    while(count < 35)
    {
		if(variant > count)
		{
			count += 5; //the stepsize
			incr ++;
		}
		else
			break;
    }

	if(incr < 2)
		return -4;
	
	if(incr == 2)
		return -3;
	
	if((incr == 3)||(incr == 6))
		return orientation*2;

	if((incr == 4)||(incr == 5))
		return orientation*1;

	if(incr == 7)
		return 3;

	return 4;
}

/* OpenGL window resize function */
void Resize(int window_width, int window_height)
{
	glViewport(0, 0, window_width, window_height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	/* set the aspect ratio */
	gluPerspective(45.0, window_width/window_height, 0.1, 100.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/* do all the OpenGL setup and config */
int SetupOpenGL()
{
	/* set smooth shading */
	glShadeModel(GL_SMOOTH);
	/* white background */
	glClearColor(1.0, 1.0, 1.0, 0.0);
	
	glClearDepth(1.0);	
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	srand((unsigned)time(NULL));
	
	size.width		= 300;
	size.height		= 150;
	g_blockNumber	= (size.width - 2)/40;
    size.width		= g_blockNumber * 40;
    g_blockVisible	= new bool[g_blockNumber * 4];
    g_paddleCoord	= (size.width - 35)/2;
    g_ballXCoord	= (size.width -3)/2;
    g_ballYCoord	= (size.height-1);

    g_changePos = 2;

	return true;
}

/* do the actual drawing */
int Render()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	glDisable(GL_BLEND);

	/* translate the entire set of blocks */
	glTranslatef(0.2, 1, -7);

	/* handle keyboard input */
	///////////////////////////
	if (g_keyboardInput[VK_LEFT])
	{
		g_paddleHorizontalCoord = -2;
	}
    
	if (g_keyboardInput[VK_RIGHT])
	{
		g_paddleHorizontalCoord = 2;
	}
	
	/* move the projectile */
	/////////////////////////
	g_ballXCoord += g_ballMoveX;
    g_ballYCoord += g_ballMoveY;
    
	if(g_ballYCoord <= 1)
    {
		g_ballMoveY =- g_ballMoveY;
		g_ballYCoord = 1;
    }

	if (g_ballXCoord <= 1)
    {
		g_ballMoveX=-g_ballMoveX;
		g_ballXCoord = 1;
    }
    
	if(g_ballYCoord >= size.height - 3)
    {
		g_ballXCoord = g_paddleCoord + 16;
		g_ballYCoord = 64;
		g_ballMoveX = 0;   
		g_ballMoveY = g_changePos;
    }
    
	if (g_ballXCoord >= (size.width - 4))
    {
		g_ballMoveX=-g_ballMoveX;
		g_ballXCoord=size.width-1-3;
    }
    
    /* paddle position */
	/////////////////////
	g_paddleCoord += g_paddleHorizontalCoord;

	if(g_paddleCoord < 1)
		g_paddleCoord = 1;
    else 
		if(g_paddleCoord > (size.width - 34))
			g_paddleCoord=(size.width - 34);
 
    if((g_ballXCoord + 3 >= g_paddleCoord)&&(g_ballYCoord >= (size.height - 5))&& 
       (g_ballYCoord < (size.height - 2))&&(g_ballXCoord <= (g_paddleCoord + 35)))
    {
		g_ballYCoord = size.height - 5;
		g_ballMoveY=-g_changePos;

		float temp_store = g_ballXCoord - g_paddleCoord;

		g_ballMoveX=ProjectilePaddleCollision(g_ballMoveX, temp_store);
    }

    /* blocks */
	////////////
	int x_coord;
	int y_coord;

    ballVelocity = g_ballMoveX;
    
    int verticalDir = g_ballMoveY;

    if(g_ballYCoord < 17 || g_ballYCoord > 64)
	{
		goto ONE;
	}

    for(int incr =  0; incr < 4; incr++)
    {
		for(int count=0; count<g_blockNumber; count++)
		{
			if(g_blockVisible[incr*g_blockNumber + count])
			{
				x_coord = 1 + count*41;
				y_coord = 20 + incr*11;
			  
				if((g_ballXCoord >= x_coord - 3) && (g_ballYCoord < y_coord + 10) &&
				   (g_ballYCoord >= y_coord - 3) && (g_ballXCoord < x_coord + 40))
				{
					g_blockVisible[incr*g_blockNumber + count] = false;
					
					/* collision detection */
					if(g_ballXCoord >= x_coord-3)
						if(g_ballXCoord <= x_coord)
						{
							g_ballMoveX=-ballVelocity;
						}

					else if(g_ballXCoord <= x_coord + 39)
						if(g_ballXCoord >= x_coord + 36)
						{
							g_ballMoveX=ballVelocity;
						}              
					g_ballMoveY =- verticalDir;
				}
			}
		}
	}

	ONE:

	/* main scene components */
	///////////////////////////
    ProjectileRenderer(g_ballXCoord, g_ballYCoord, 3, 3);
	ProjectileRenderer(g_paddleCoord, size.height - 2, 35, 2);
    
	/* blocks */
	////////////
    int blockVisible = 1;

	for(int incr = 0; incr < 4; incr++)
    {
		for (int count = 0; count < g_blockNumber; count++)
		{
			if (g_blockVisible[incr*g_blockNumber + count])
			{
				blockVisible = 0;
				BlockRenderer(1 + count*41, 20 + incr*11, 40, 10);
			}
		}
    }
    if(blockVisible == 1)
    {
		for(int count = 0; count < 4*g_blockNumber; count++)
			g_blockVisible[count]=true;
    }

	return true;
}

BOOL MainWindow(char* caption, int window_width, int window_height, int colour_depth)
{
	int	pixelStorage;
	
	RECT rectangleStruct;
	rectangleStruct.top= 0;
	rectangleStruct.bottom = window_height;
	rectangleStruct.left = 0;
	rectangleStruct.right= window_width;

	hInstance = GetModuleHandle(NULL);
	windowClassStruct.hCursor = LoadCursor(NULL, IDC_ARROW);
	windowClassStruct.hbrBackground	= NULL;
	windowClassStruct.hInstance = hInstance;
	windowClassStruct.hIcon = NULL;
	windowClassStruct.style = CS_HREDRAW | CS_VREDRAW;
	windowClassStruct.lpfnWndProc = (WNDPROC)WndProc;
	windowClassStruct.lpszMenuName = NULL;
	windowClassStruct.cbClsExtra  = 0;
	windowClassStruct.cbWndExtra = 0;
	windowClassStruct.lpszClassName	= "Breakout";

	/* register the window class */
	RegisterClass(&windowClassStruct);
	
	dwWindowStyle=WS_OVERLAPPEDWINDOW;
	dwExtendedStyle = WS_EX_APPWINDOW;

	AdjustWindowRectEx(&rectangleStruct, dwWindowStyle, FALSE, dwExtendedStyle);

	/* window creation */
	hWindow = CreateWindowEx(dwExtendedStyle, 
							 "Breakout",
							 caption,
							 dwWindowStyle,	0, 0,
							 rectangleStruct.right-rectangleStruct.left,
							 rectangleStruct.bottom-rectangleStruct.top,
							 NULL, NULL, hInstance, NULL);
		
	hDeviceContext = GetDC(hWindow);

	/* set the pixel format descriptor */
	static	PIXELFORMATDESCRIPTOR mainDescriptor=
	{
		sizeof(PIXELFORMATDESCRIPTOR), 2,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL |PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA, colour_depth, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 16, 0, 0, PFD_MAIN_PLANE, 0, 0, 0, 0 };


	pixelStorage = ChoosePixelFormat(hDeviceContext,&mainDescriptor);

	SetPixelFormat(hDeviceContext,pixelStorage,&mainDescriptor);
	hRenderContext=wglCreateContext(hDeviceContext);
	wglMakeCurrent(hDeviceContext, hRenderContext);

	/* display the window */
	ShowWindow(hWindow,SW_SHOW);
	SetForegroundWindow(hWindow);
	
	/* keyboard focus */
	SetFocus(hWindow); 

	Resize(window_width, window_height);

	SetupOpenGL();

	return true;
}

LRESULT CALLBACK WndProc(HWND hWindow, UINT	msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
		case WM_CREATE:
			SetTimer(hWindow, 1, 20, NULL);
			return 0;
		
		case WM_CLOSE:
		{
			PostQuitMessage(0);
			return 0;
		}

		case WM_KEYDOWN:
		{
			g_keyboardInput[wParam] = TRUE;

			return 0;
		}

		case WM_KEYUP:
		{
			g_keyboardInput[wParam] = FALSE;
			return 0;
		}

		case WM_SIZE:
		{
			Resize(LOWORD(lParam),HIWORD(lParam));
			return 0;
		}
		case WM_TIMER:
			if (g_programRunning == 1)
			{
				Render();
				SwapBuffers(hDeviceContext);
			}
			return 0;
	}

	/* deal with all unhandled msgs */
	return DefWindowProc(hWindow,msg,wParam,lParam);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG	msg;

	MainWindow("Breakout Clone", 800, 600, 32);

	SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);

	while(GetMessage(&msg, NULL, 0, 0)) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	delete[] g_blockVisible;
	return (msg.wParam);
}
